#!/usr/bin/python
# coding=utf-8

from __future__ import division, print_function, unicode_literals

__all__ = ["__version__", "__author__", "__author_email__"]

__version__ = "0.5"

__author__ = 'Klaus Greff'
__author_email__ = 'qwlouse@gmail.com',

__url__ = "https://github.com/Qwlouse/sacred"



#!/usr/bin/env python
# coding=utf-8
from __future__ import division, print_function, unicode_literals
from sacred.utils import join_paths


__sacred__ = True  # marker for filtering stacktraces when run from commandline


class FallbackDict(dict):
    """
    This dictionary either returns the value assigned to a given key or it
    returns the value for that key from the fallback dict.
    """
    def __init__(self, fallback, **kwargs):
        super(FallbackDict, self).__init__(**kwargs)
        self.fallback = fallback

    def __getitem__(self, item):
        if dict.__contains__(self, item):
            return dict.__getitem__(self, item)
        else:
            return self.fallback[item]

    def __contains__(self, item):
        return dict.__contains__(self, item) or (item in self.fallback)

    def get(self, k, d=None):
        if dict.__contains__(self, k):
            return dict.__getitem__(self, k)()
        else:
            return self.fallback.get(k, d)

    def has_key(self, item):
        return self.__contains__(item)

    def items(self):
        raise NotImplemented

    def iteritems(self):
        raise NotImplemented

    def iterkeys(self):
        raise NotImplemented

    def itervalues(self):
        raise NotImplemented

    def keys(self):
        raise NotImplemented

    def pop(self, k, d=None):
        raise NotImplemented

    def popitem(self):
        raise NotImplemented

    def setdefault(self, k, d=None):
        raise NotImplemented

    def update(self, E=None, **F):
        raise NotImplemented

    def values(self):
        raise NotImplemented

    def viewitems(self):
        raise NotImplemented

    def viewkeys(self):
        raise NotImplemented

    def viewvalues(self):
        raise NotImplemented

    def __iter__(self):
        raise NotImplemented

    def __len__(self):
        raise NotImplemented


class DogmaticDict(dict):
    def __init__(self, fixed=None, fallback=None):
        super(DogmaticDict, self).__init__()
        self.typechanges = {}
        self._fixed = fixed or {}
        self._fallback = {}
        if fallback:
            self.fallback = fallback

    @property
    def fallback(self):
        return self._fallback

    @fallback.setter
    def fallback(self, newval):
        ffkeys = set(self._fixed.keys()).intersection(set(newval.keys()))
        if ffkeys:
            raise ValueError("Keys %s appear are both fixed and fallback keys"
                             % ffkeys)
        self._fallback = newval

    def __setitem__(self, key, value):
        if key not in self._fixed:
            dict.__setitem__(self, key, value)
        else:
            fixed_val = self._fixed[key]
            dict.__setitem__(self, key, fixed_val)
            # log typechanges
            if type_changed(value, fixed_val):
                self.typechanges[key] = (type(value), type(fixed_val))

            if isinstance(fixed_val, DogmaticDict) and isinstance(value, dict):
                # recursive update
                for k, v in value.items():
                    fixed_val[k] = v

                for k, v in fixed_val.typechanges.items():
                    self.typechanges[join_paths(key, k)] = v

    def __getitem__(self, item):
        if dict.__contains__(self, item):
            return dict.__getitem__(self, item)
        elif item in self.fallback:
            if item in self._fixed:
                return self._fixed[item]
            else:
                return self.fallback[item]
        raise KeyError(item)

    def __contains__(self, item):
        return dict.__contains__(self, item) or (item in self.fallback)

    def get(self, k, d=None):
        if dict.__contains__(self, k):
            return dict.__getitem__(self, k)()
        else:
            return self.fallback.get(k, d)

    def has_key(self, item):
        return self.__contains__(item)

    def __delitem__(self, key):
        if key not in self._fixed:
            dict.__delitem__(self, key)

    def update(self, iterable=None, **kwargs):
        if iterable is not None:
            if hasattr(iterable, 'keys'):
                for k in iterable:
                    self[k] = iterable[k]
            else:
                for (k, v) in iterable:
                    self[k] = v
        for k in kwargs:
            self[k] = kwargs[k]

    def revelation(self):
        missing = set()
        for key in self._fixed:
            if not dict.__contains__(self, key):
                self[key] = self._fixed[key]
                missing.add(key)

            if isinstance(self[key], (DogmaticDict, DogmaticList)):
                missing |= {key + "." + k for k in self[key].revelation()}
        return missing


class DogmaticList(list):
    def append(self, p_object):
        pass

    def extend(self, iterable):
        pass

    def insert(self, index, p_object):
        pass

    def reverse(self):
        pass

    def sort(self, compare=None, key=None, reverse=False):
        pass

    def __iadd__(self, other):
        return self

    def __imul__(self, other):
        return self

    def __setitem__(self, key, value):
        pass

    def __setslice__(self, i, j, sequence):
        pass

    def __delitem__(self, key):
        pass

    def __delslice__(self, i, j):
        pass

    def pop(self, index=None):
        raise TypeError('Cannot pop from DogmaticList')

    def remove(self, value):
        pass

    def revelation(self):
        for i, x in enumerate(self):
            if isinstance(x, (DogmaticDict, DogmaticList)):
                x.revelation()
        return set()


simplify_type = {
    bool: bool,
    float: float,
    int: int,
    str: str,
    list: list,
    tuple: list,
    dict: dict,
    DogmaticDict: dict,
    DogmaticList: list,
}

# if in python 2 we want to ignore unicode/str and int/long typechanges
try:
    simplify_type[unicode] = str
    simplify_type[long] = int
except NameError:
    pass

# if numpy is available we also want to ignore typechanges from numpy
# datatypes to the corresponding python datatype
try:
    import numpy as np
    np_floats = [np.float, np.float16, np.float32, np.float64, np.float128]
    for npf in np_floats:
        simplify_type[npf] = float

    np_ints = [np.int, np.int8, np.int16, np.int32, np.int64,
               np.uint8, np.uint16, np.uint32, np.uint64]
    for npi in np_ints:
        simplify_type[npi] = int

    simplify_type[np.bool_] = bool
except ImportError:
    pass


def type_changed(a, b):
    return simplify_type[type(a)] != simplify_type[type(b)]


def dogmatize(x):
    if isinstance(x, dict):
        return DogmaticDict({k: dogmatize(v) for k, v in x.items()})
    elif isinstance(x, list):
        return DogmaticList([dogmatize(v) for v in x])
    elif isinstance(x, tuple):
        return tuple(dogmatize(v) for v in x)
    else:
        return x


def undogmatize(x):
    if isinstance(x, DogmaticDict):
        return dict({k: undogmatize(v) for k, v in x.items()})
    elif isinstance(x, DogmaticList):
        return list([undogmatize(v) for v in x])
    elif isinstance(x, tuple):
        return tuple(undogmatize(v) for v in x)
    else:
        return x



#!/usr/bin/env python
# coding=utf-8

from __future__ import division, print_function, unicode_literals
from datetime import datetime
import pickle

__all__ = ['RunObserver', 'MongoObserver', 'DebugObserver']


class RunObserver(object):
    def started_event(self, name, ex_info, host_info, start_time, config):
        pass

    def heartbeat_event(self, info, captured_out):
        pass

    def completed_event(self, stop_time, result):
        pass

    def interrupted_event(self, interrupt_time):
        pass

    def failed_event(self, fail_time, fail_trace):
        pass


try:
    from pymongo import MongoClient
    from pymongo.son_manipulator import SONManipulator
    from bson import Binary

    SON_MANIPULATORS = []

    try:
        import numpy as np

        class PickleNumpyArrays(SONManipulator):
            """
            Helper that makes sure numpy arrays get pickled and stored in the
            database as binary strings.
            """
            def transform_incoming(self, son, collection):
                for (key, value) in son.items():
                    if isinstance(value, np.ndarray):
                        son[key] = {
                            "_type": "ndarray",
                            "_value": Binary(pickle.dumps(value, protocol=2))}
                    elif isinstance(value, dict):
                        # Make sure we recurse into sub-docs
                        son[key] = self.transform_incoming(value, collection)
                return son

            def transform_outgoing(self, son, collection):
                for (key, value) in son.items():
                    if isinstance(value, dict):
                        if "_type" in value and value["_type"] == "ndarray":
                            son[key] = pickle.loads(str(value["_value"]))
                        else:  # Again, make sure to recurse into sub-docs
                            son[key] = self.transform_outgoing(value,
                                                               collection)
                return son

        SON_MANIPULATORS.append(PickleNumpyArrays())
    except ImportError:
        pass

    class MongoObserver(RunObserver):
        def __init__(self, url=None, db_name='sacred'):
            RunObserver.__init__(self)
            self.experiment_entry = None
            mongo = MongoClient(url)
            self.db = mongo[db_name]
            for manipulator in SON_MANIPULATORS:
                self.db.add_son_manipulator(manipulator)
            self.collection = self.db['experiments']

        def save(self):
            self.collection.save(self.experiment_entry)

        def started_event(self, name, ex_info, host_info, start_time, config):
            self.experiment_entry = dict()
            self.experiment_entry['name'] = name
            self.experiment_entry['experiment_info'] = ex_info
            try:
                with open(ex_info['mainfile'], 'r') as f:
                    self.experiment_entry['source'] = f.read()
            except IOError as e:
                self.experiment_entry['experiment_info']['source'] = str(e)
            self.experiment_entry['host_info'] = host_info
            self.experiment_entry['start_time'] = \
                datetime.fromtimestamp(start_time)
            self.experiment_entry['config'] = config
            self.experiment_entry['status'] = 'RUNNING'
            self.save()

        def heartbeat_event(self, info, captured_out):
            self.experiment_entry['info'] = info
            self.experiment_entry['captured_out'] = captured_out
            self.experiment_entry['heartbeat'] = datetime.now()
            self.save()

        def completed_event(self, stop_time, result):
            self.experiment_entry['stop_time'] = \
                datetime.fromtimestamp(stop_time)
            self.experiment_entry['result'] = result
            self.experiment_entry['status'] = 'COMPLETED'
            self.save()

        def interrupted_event(self, interrupt_time):
            self.experiment_entry['stop_time'] = \
                datetime.fromtimestamp(interrupt_time)
            self.experiment_entry['status'] = 'INTERRUPTED'
            self.save()

        def failed_event(self, fail_time, fail_trace):
            self.experiment_entry['stop_time'] = \
                datetime.fromtimestamp(fail_time)
            self.experiment_entry['status'] = 'FAILED'
            self.experiment_entry['fail_trace'] = fail_trace
            self.save()

        def __eq__(self, other):
            if isinstance(other, MongoObserver):
                return self.collection == other.collection
            return False

        def __ne__(self, other):
            return not self.__eq__(other)

except ImportError:
    class MongoObserver(RunObserver):
        def __init__(self, url=None, db_name='sacred'):
            raise ImportError('only available if "pymongo" is installed')


class DebugObserver(RunObserver):
    def started_event(self, name, ex_info, host_info, start_time, config):
        print('experiment_started_event')

    def heartbeat_event(self, info, captured_out):
        print('experiment_info_updated')

    def completed_event(self, stop_time, result):
        print('experiment_completed_event')

    def interrupted_event(self, interrupt_time):
        print('experiment_interrupted_event')

    def failed_event(self, fail_time, fail_trace):
        print('experiment_failed_event')



#!/usr/bin/python
# coding=utf-8

from __future__ import division, print_function, unicode_literals
from collections import OrderedDict, defaultdict, namedtuple
from copy import copy
from sacred.config_scope import chain_evaluate_config_scopes
from sacred.custom_containers import dogmatize
from sacred.host_info import get_host_info
from sacred.run import Run
from sacred.utils import (
    get_seed, create_rnd, is_prefix, set_by_dotted_path, iterate_flattened,
    iter_prefixes, iter_path_splits, create_basic_stream_logger, join_paths,
    get_by_dotted_path, convert_to_nested_dict)


__sacred__ = True  # marker for filtering stacktraces when run from commandline


class Scaffold(object):
    def __init__(self, config_scopes, subrunners, path, captured_functions,
                 commands, named_configs, generate_seed):
        self.config_scopes = config_scopes
        self.named_configs = named_configs
        self.subrunners = subrunners
        self.path = path
        self.generate_seed = generate_seed
        self.config_updates = {}
        self.named_configs_to_use = []
        self.config = None
        self.fixture = None  # TODO: rename
        self.logger = None
        self.seed = None
        self.rnd = None
        self._captured_functions = captured_functions
        self.commands = commands

    def set_up_seed(self, rnd=None):
        if self.seed is not None:
            return

        self.seed = self.config_updates.get('seed') or get_seed(rnd)
        self.rnd = create_rnd(self.seed)

        # Hierarchically set the seed of proper subrunners
        for subrunner_path, subrunner in reversed(list(
                self.subrunners.items())):
            if is_prefix(self.path, subrunner_path):
                subrunner.set_up_seed(self.rnd)

    def set_up_config(self):
        if self.config is not None:
            return self.config

        # gather presets
        fallback = {}
        for sr_path, sr in self.subrunners.items():
            if self.path and is_prefix(self.path, sr_path):
                path = sr_path[len(self.path):].strip('.')
                set_by_dotted_path(fallback, path, sr.config)
            else:
                set_by_dotted_path(fallback, sr_path, sr.config)

        # dogmatize to make the subrunner configurations read-only
        const_fallback = dogmatize(fallback)
        const_fallback.revelation()

        self.config = {}

        if self.generate_seed:
            self.config['seed'] = self.seed

        # named configs first
        self.config_updates = chain_evaluate_config_scopes(
            [self.named_configs[n] for n in self.named_configs_to_use],
            fixed=self.config_updates,
            preset=self.config,
            fallback=const_fallback)

        # unnamed (default) configs second
        self.config = chain_evaluate_config_scopes(
            self.config_scopes,
            fixed=self.config_updates,
            preset=self.config,
            fallback=const_fallback)

        return self.config

    def get_config_modifications(self):
        typechanges = {}
        flat_config_upd = [k
                           for k, v in iterate_flattened(self.config_updates)]
        updated = {sp for p in flat_config_upd for sp in iter_prefixes(p)}
        added = set(updated)
        for config in self.config_scopes:
            added &= config.added_values
            typechanges.update(config.typechanges)

        if self.generate_seed and 'seed' in added:
            added.remove('seed')

        return added, updated, typechanges

    def get_fixture(self):
        if self.fixture is not None:
            return self.fixture

        self.fixture = copy(self.config)
        for sr_path, sr in self.subrunners.items():
            sub_fix = sr.get_fixture()
            sub_path = sr_path
            if is_prefix(self.path, sub_path):
                sub_path = sr_path[len(self.path):].strip('.')
            # Note: This might fail if we allow non-dict fixtures
            set_by_dotted_path(self.fixture, sub_path, sub_fix)
        return self.fixture

    def finalize_initialization(self, run):
        # look at seed again, because it might have changed during the
        # configuration process
        if 'seed' in self.config:
            self.seed = self.config['seed']
            self.rnd = create_rnd(self.seed)

        for cf in self._captured_functions:
            cf.logger = self.logger.getChild(cf.__name__)
            cf.config = get_by_dotted_path(self.get_fixture(), cf.prefix)
            seed = get_seed(self.rnd)
            cf.rnd = create_rnd(seed)
            cf.run = run

        self._warn_about_suspicious_changes()

    def _warn_about_suspicious_changes(self, ):
        add, upd, tch = self.get_config_modifications()
        for a in sorted(add):
            self.logger.warning('Added new config entry: "%s"' % a)
        for k, (t1, t2) in tch.items():
            if (isinstance(t1, type(None)) or
                    (t1 in (int, float) and t2 in (int, float))):
                continue
            self.logger.warning(
                'Changed type of config entry "%s" from %s to %s' %
                (k, t1.__name__, t2.__name__))


def get_configuration(scaffolding):
    config = {}
    for sc_path, sc in reversed(list(scaffolding.items())):
        if sc_path:
            set_by_dotted_path(config, sc_path, sc.config)
        else:
            config.update(sc.config)
    return config


def distribute_config_updates(scaffolding, config_updates):
    if config_updates is None:
        return
    nested_config_updates = convert_to_nested_dict(config_updates)
    for path, value in iterate_flattened(nested_config_updates):
        for p1, p2 in reversed(list(iter_path_splits(path))):
            if p1 in scaffolding:
                set_by_dotted_path(scaffolding[p1].config_updates, p2, value)
                break
                # this is guaranteed to occur for one of the modrunners,
                # because the exrunner has path ''


def distribute_named_configs(scaffolding, named_configs):
    for ncfg in named_configs:
        path, _, cfg_name = ncfg.rpartition('.')
        if path not in scaffolding:
            raise KeyError('Ingredient for named config "%s" not found' % ncfg)
        scaffolding[path].named_configs_to_use.append(cfg_name)


def initialize_logging(experiment, scaffolding, loglevel=None):
    if experiment.logger is None:
        if loglevel:
            try:
                loglevel = int(loglevel)
            except ValueError:
                pass
        root_logger = create_basic_stream_logger('', level=loglevel)
    else:
        root_logger = experiment.logger
        if loglevel:
            root_logger.setLevel(loglevel)

    for sc_path, sc in scaffolding.items():
        if sc_path:
            sc.logger = root_logger.getChild(sc_path)
        else:
            sc.logger = root_logger

    return root_logger.getChild(experiment.name)


def create_scaffolding(experiment):
    sorted_ingredients = gather_ingredients_topological(experiment)
    scaffolding = OrderedDict()
    for sm in sorted_ingredients:
        scaffolding[sm] = Scaffold(
            sm.cfgs,
            subrunners=OrderedDict([(scaffolding[m].path, scaffolding[m])
                                    for m in sm.ingredients]),
            path=sm.path if sm != experiment else '',
            captured_functions=sm.captured_functions,
            commands=sm.commands,
            named_configs=sm.named_configs,
            generate_seed=sm.gen_seed)
    return OrderedDict([(sc.path, sc) for sc in scaffolding.values()])


def gather_ingredients_topological(ingredient):
    sub_ingredients = defaultdict(int)
    for sm, depth in ingredient.traverse_ingredients():
        sub_ingredients[sm] = max(sub_ingredients[sm], depth)
    return sorted(sub_ingredients, key=lambda x: -sub_ingredients[x])


ConfigModifications = namedtuple('ConfigModifications',
                                 ['added', 'updated', 'typechanges'])


def get_config_modifications(scaffolding):
    added = set()
    updated = set()
    typechanges = {}
    for sc_path, sc in scaffolding.items():
        mr_add, mr_up, mr_tc = sc.get_config_modifications()
        if mr_add or mr_up or mr_tc:
            updated |= set(iter_prefixes(sc_path))
        added |= {join_paths(sc_path, a) for a in mr_add}
        updated |= {join_paths(sc_path, u) for u in mr_up}
        typechanges.update({join_paths(sc_path, k): v
                            for k, v in mr_tc.items()})
    return ConfigModifications(added, updated, typechanges)


def get_command(scaffolding, command_path):
    path, _, command_name = command_path.rpartition('.')
    if path not in scaffolding:
        raise KeyError('Ingredient for command "%s" not found.' % command_path)

    if command_name in scaffolding[path].commands:
        return scaffolding[path].commands[command_name]
    else:
        if path:
            raise KeyError('Command "%s" not found in ingredient "%s"' %
                           (command_name, path))
        else:
            raise KeyError('Command "%s" not found' % command_name)


def create_run(experiment, command_name, config_updates=None, log_level=None,
               named_configs=()):
    scaffolding = create_scaffolding(experiment)
    logger = initialize_logging(experiment, scaffolding, log_level)

    distribute_config_updates(scaffolding, config_updates)
    distribute_named_configs(scaffolding, named_configs)

    for sc in reversed(list(scaffolding.values())):
        sc.set_up_seed()  # partially recursive

    for sc in scaffolding.values():
        sc.set_up_config()

    config = get_configuration(scaffolding)

    config_modifications = get_config_modifications(scaffolding)

    # only get experiment and host info if there are observers
    if experiment.observers:
        experiment_info = experiment.get_info()
        host_info = get_host_info()
    else:
        experiment_info = host_info = dict()

    main_function = get_command(scaffolding, command_name)

    run = Run(config, config_modifications, main_function,
              experiment.observers, logger, experiment.name, experiment_info,
              host_info)

    for sc in scaffolding.values():
        sc.finalize_initialization(run=run)

    return run



#!/usr/bin/env python
# coding=utf-8

from __future__ import division, print_function, unicode_literals
from collections import OrderedDict
import inspect
import os.path
import sys

from sacred.arg_parser import get_config_updates, get_observers, parse_args
from sacred.captured_function import create_captured_function
from sacred.commands import print_config
from sacred.config_scope import ConfigScope
from sacred.host_info import get_dependencies, fill_missing_versions
from sacred.initialize import create_run
from sacred.utils import print_filtered_stacktrace


__sacred__ = True  # marker for filtering stacktraces when run from commandline


class CircularDependencyError(Exception):
    pass


class Ingredient(object):
    def __init__(self, path, ingredients=(), gen_seed=False,
                 caller_globals=None):
        self.path = path
        self.cfgs = []
        self.named_configs = dict()
        self.ingredients = list(ingredients)
        self.gen_seed = gen_seed
        self.captured_functions = []
        self._is_traversing = False
        self.commands = OrderedDict()
        # capture some context information
        caller_globals = caller_globals or inspect.stack()[1][0].f_globals
        self.doc = caller_globals.get('__doc__') or ""
        self.mainfile = caller_globals.get('__file__') or ""
        if self.mainfile:
            self.mainfile = os.path.abspath(self.mainfile)
            if self.mainfile.endswith('.pyc'):
                non_compiled_mainfile = self.mainfile[:-1]
                if os.path.exists(non_compiled_mainfile):
                    self.mainfile = non_compiled_mainfile

        self.dependencies = get_dependencies(caller_globals)

    # =========================== Decorators ==================================
    def command(self, func=None, prefix=None):
        """
        Decorator to define a new command for this Ingredient or Experiment.

        The name of the command will be the name of the function. It can be
        called from the commandline or by using the run_command function.

        Commands are automatically also captured functions.
        """
        def _command(f):
            captured_f = self.capture(f, prefix=prefix)
            self.commands[f.__name__] = captured_f
            return captured_f

        if func is not None:
            return _command(func)
        else:
            return _command

    def config(self, func):
        """
        Decorator to turn a function into a ConfigScope and add it to the
        Ingredient/Experiment.
        """
        self.cfgs.append(ConfigScope(func))
        return self.cfgs[-1]

    def named_config(self, func):
        config_scope = ConfigScope(func)
        self.named_configs[func.__name__] = config_scope
        return config_scope

    def capture(self, func=None, prefix=None):
        """
        Decorator to turn a function into a captured function.
        """
        def _capture(f):
            if f in self.captured_functions:
                return f
            captured_function = create_captured_function(f, prefix=prefix)
            self.captured_functions.append(captured_function)
            return captured_function

        if func is not None:
            return _capture(func)
        else:
            return _capture

    # ======================== protected helpers ==============================
    def traverse_ingredients(self):
        if self._is_traversing:
            raise CircularDependencyError()
        else:
            self._is_traversing = True
        yield self, 0
        for ingredient in self.ingredients:
            for sr, depth in ingredient.traverse_ingredients():
                yield sr, depth + 1
        self._is_traversing = False

    def run_command(self, command_name, config_updates=None,
                    named_configs_to_use=(), loglevel=None):
        run = create_run(self, command_name, config_updates,
                         log_level=loglevel,
                         named_configs=named_configs_to_use)
        run.logger.info("Running command '%s'" % command_name)
        return run()

    def _gather_commands(self):
        for k, v in self.commands.items():
            yield self.path + '.' + k, v

        for ingred in self.ingredients:
            for k, v in ingred._gather_commands():
                yield k, v


class Experiment(Ingredient):
    def __init__(self, name, ingredients=()):
        caller_globals = inspect.stack()[1][0].f_globals
        super(Experiment, self).__init__(path=name,
                                         ingredients=ingredients,
                                         gen_seed=True,
                                         caller_globals=caller_globals)
        self.name = name
        self.default_command = None
        self.logger = None
        self.observers = []
        self.command(print_config)
        self.info = None

    # =========================== Decorators ==================================

    def main(self, f):
        """
        Decorator to define the main function of the experiment.

        The main function of an experiment is the default command that is being
        run when no command is specified, or when calling the run() method.
        """
        captured = self.command(f)
        self.default_command = captured.__name__
        return captured

    def automain(self, f):
        """
        Decorator that defines the main function of the experiment and
        automatically runs the experiment commandline when the file is
        executed.

        The method decorated by this should be last in the file because:

        .. code-block:: python

            @ex.automain
            def my_main():
                pass

        is equivalent to:

        .. code-block:: python

            @ex.main
            def my_main():
                pass

            if __name__ == '__main__':
                ex.run_commandline()
        """
        captured = self.main(f)
        if f.__module__ == '__main__':
            self.run_commandline()
        return captured

    # =========================== public interface ============================

    def get_info(self):
        fill_missing_versions(self.dependencies)

        return dict(
            mainfile=self.mainfile,
            dependencies=self.dependencies.items(),
            doc=self.doc)

    def run(self, config_updates=None, named_configs=(), loglevel=None):
        """
        Run the main function of the experiment.

        :param config_updates: Changes to the configuration as a nested
                               dictionary
        :type config_updates: dict
        :param named_configs: list of names of named_configs to use
        :type named_configs: list
        :param loglevel: Changes to the log-level for this run.
        :type loglevel: int | str

        :return: The result of the main function.
        """
        return self.run_command(self.default_command,
                                config_updates=config_updates,
                                named_configs_to_use=named_configs,
                                loglevel=loglevel)

    def run_commandline(self, argv=None):
        if argv is None:
            argv = sys.argv
        all_commands = self._gather_commands()

        args = parse_args(argv,
                          description=self.doc,
                          commands=OrderedDict(all_commands),
                          print_help=True)
        config_updates, named_configs = get_config_updates(args['UPDATE'])
        loglevel = args.get('--logging')
        for obs in get_observers(args):
            if obs not in self.observers:
                self.observers.append(obs)

        if args['COMMAND']:
            cmd_name = args['COMMAND']
        else:
            cmd_name = self.default_command

        try:
            return self.run_command(cmd_name,
                                    config_updates=config_updates,
                                    named_configs_to_use=named_configs,
                                    loglevel=loglevel)
        except:
            if args['--debug']:
                raise
            else:
                print_filtered_stacktrace()

    # ============================ protected interface ========================

    def _gather_commands(self):
        for k, v in self.commands.items():
            yield k, v

        for ingred in self.ingredients:
            for k, v in ingred._gather_commands():
                yield k, v



#!/usr/bin/env python
# coding=utf-8

from __future__ import division, print_function, unicode_literals
import pprint
import pydoc

from sacred.utils import iterate_separately, join_paths


BLUE = '\033[94m'
GREEN = '\033[92m'
RED = '\033[91m'
ENDC = '\033[0m'


def _my_safe_repr(objekt, context, maxlevels, level):
    """
    Used to override the pprint format method in order to get rid of
    unnecessary unicode prefixes. E.g.: 'John' instead of u'John'.
    """
    typ = pprint._type(objekt)
    if typ is unicode:
        try:
            objekt = str(objekt)
        except UnicodeEncodeError:
            pass
    return pprint._safe_repr(objekt, context, maxlevels, level)


def _cfgprint(x, key, added, updated, typechanges, indent=''):
    def colored(text):
        if key in added:
            return GREEN + text + ENDC
        elif key in typechanges:
            return RED + text + ENDC
        elif key in updated:
            return BLUE + text + ENDC
        else:
            return text

    last_key = key.split('.')[-1]
    if isinstance(x, dict):
        if last_key:
            print(colored('{}{}:'.format(indent, last_key)))
        for k, v in iterate_separately(x):
            subkey = join_paths(key, k)
            _cfgprint(v, subkey, added, updated, typechanges, indent + '  ')
    else:
        printer = pprint.PrettyPrinter(indent=len(indent)+2)
        printer.format = _my_safe_repr
        print(colored('{}{} = {}'.format(indent, last_key,
                                         printer.pformat(x))))

LEGEND = '(' + BLUE + 'modified' + ENDC +\
    ', ' + GREEN + 'added' + ENDC +\
    ', ' + RED + 'typechanged' + ENDC + ')'


def print_config(_run):
    """
    Print the updated configuration and exit.

    Text is highlighted:
      green:  value updated
      blue:   value added
      red:    value updated but type changed
    """
    final_config = _run.config
    added, updated, typechanges = _run.config_modifications
    print('Configuration', LEGEND + ':')
    _cfgprint(final_config, '', added, updated, typechanges)


def help_for_command(command):
    return pydoc.text.document(command)



#!/usr/bin/env python
# coding=utf-8
from __future__ import division, print_function, unicode_literals
from datetime import timedelta
import time

import wrapt

from sacred.custom_containers import FallbackDict
from sacred.signature import Signature
from sacred.utils import create_rnd, get_seed

__sacred__ = True


def create_captured_function(f, prefix=None):
    f.signature = Signature(f)
    f.logger = None
    f.config = {}
    f.rnd = None
    f.run = None
    f.prefix = prefix
    return captured_function(f)


@wrapt.decorator
def captured_function(wrapped, instance, args, kwargs):
    # todo: performance optimize this by only creating a PRNG if the signature
    # contains either _seed or _rnd
    runseed = get_seed(wrapped.rnd)
    options = FallbackDict(
        wrapped.config,
        _log=wrapped.logger,
        _seed=runseed,
        _rnd=create_rnd(runseed),
        _run=wrapped.run
    )
    args, kwargs = wrapped.signature.construct_arguments(args, kwargs, options)
    wrapped.logger.debug("Started")
    start_time = time.time()
    # =================== run actual function =================================
    result = wrapped(*args, **kwargs)
    # =========================================================================
    stop_time = time.time()
    elapsed_time = timedelta(seconds=round(stop_time - start_time))
    wrapped.logger.debug("Finished after %s." % elapsed_time)

    return result



#!/usr/bin/env python
# coding=utf-8
from __future__ import division, print_function, unicode_literals
import multiprocessing
import platform
import re
import subprocess
import sys
import pkg_resources

try:
    basestring  # attempt to evaluate basestring

    def is_str(s):
        return isinstance(s, basestring)
except NameError:
    def is_str(s):
        return isinstance(s, str)


def get_processor_name():
    if platform.system() == "Windows":
        return platform.processor().strip()
    elif platform.system() == "Darwin":
        import os
        os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
        command = "sysctl -n machdep.cpu.brand_string"
        return subprocess.check_output(command).strip()
    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = str(subprocess.check_output(command, shell=True)).strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                return re.sub(".*model name.*:", "", line, 1).strip()
    return ""


def get_host_info():
    return {
        "cpu": get_processor_name(),
        "cpu_count": multiprocessing.cpu_count(),
        "hostname": platform.node(),
        "os": platform.system(),
        "os_info": platform.platform(),
        "python_version": platform.python_version(),
        "python_compiler": platform.python_compiler()
    }

MODULE_BLACKLIST = {None, '__future__'} | set(sys.builtin_module_names)
module = type(platform)


def get_dependencies(globs):
    dependencies = {}

    for g in globs.values():
        if isinstance(g, module) and g.__name__ not in MODULE_BLACKLIST:
            dependencies[g.__name__] = get_module_version_heuristic(g)

        elif hasattr(g, '__module__'):
            modname = g.__module__.split('.')[0]
            if modname not in MODULE_BLACKLIST and modname not in dependencies:
                dependencies[modname] = get_module_version_heuristic(modname)

    return dependencies


def fill_missing_versions(deps):
    for mod_name, ver in deps.items():
        if ver is None:
            deps[mod_name] = get_module_version_from_pkg_resources(mod_name)


def get_modules(globs):
    return {g for g in globs.values()
            if isinstance(g, module) and g.__name__ not in MODULE_BLACKLIST}


PEP440_VERSION_PATTERN = re.compile(r"""
^
(\d+!)?              # epoch
(\d[\.\d]*(?<= \d))  # release
((?:[abc]|rc)\d+)?   # pre-release
(?:(\.post\d+))?     # post-release
(?:(\.dev\d+))?      # development release
$
""", flags=re.VERBOSE)


def get_module_version_heuristic(mod):
    if not isinstance(mod, module):
        mod = sys.modules.get(mod)
        if not mod:
            return None
    possible_version_attributes = {'version', 'VERSION', '__version__'}
    for vattr in possible_version_attributes:
        if not hasattr(mod, vattr):
            continue
        v = getattr(mod, vattr)
        if is_str(v) and PEP440_VERSION_PATTERN.match(v):
            return v
    return None


def get_module_version_from_pkg_resources(mod_name):
    try:
        return pkg_resources.get_distribution(mod_name).version
    except pkg_resources.DistributionNotFound:
        return None



#!/usr/bin/env python
# coding=utf-8
from __future__ import division, print_function, unicode_literals
from collections import OrderedDict
from copy import copy
import inspect


__sacred__ = True  # marker for filtering stacktraces when run from commandline


class Signature:
    """
    Contains information about the signature of a function.
    name : the functions name
    arguments : list of all arguments
    vararg_name : name of the *args variable
    kw_wildcard_name : name of the **kwargs variable
    positional_args : list of all positional-only arguments
    kwargs : dict of all keyword arguments mapped to their default
    """
    def __init__(self, f):
        self.name = f.__name__
        args, vararg_name, kw_wildcard_name, defaults = inspect.getargspec(f)
        self.arguments = args
        self.vararg_name = vararg_name
        self.kw_wildcard_name = kw_wildcard_name
        defaults = defaults or []
        self.positional_args = args[:len(args) - len(defaults)]
        self.kwargs = OrderedDict(zip(args[-len(defaults):], defaults))

    def get_free_parameters(self, args, kwargs):
        return [a for a in self.arguments[len(args):] if a not in kwargs]

    def construct_arguments(self, args, kwargs, options):
        """
        Construct args list and kwargs dictionary for this signature such that:
          - the original explicit call arguments (args, kwargs) are preserved
          - missing arguments are filled in by name using options (if possible)
          - default arguments are overridden by options
          - TypeError is thrown if:
            * kwargs contains one or more unexpected keyword arguments
            * conflicting values for a parameter in both args and kwargs
            * there is an unfilled parameter at the end of this process
        """
        self._assert_no_unexpected_args(args)
        self._assert_no_unexpected_kwargs(kwargs)
        self._assert_no_duplicate_args(args, kwargs)
        if self.kw_wildcard_name:
            k = copy(options)
            k.update(kwargs)
            kwargs = k
        else:
            args, kwargs = self._fill_in_options(args, kwargs, options)

        self._assert_no_missing_args(args, kwargs)
        return args, kwargs

    def __unicode__(self):
        args = self.positional_args
        vararg = ("*" + self.vararg_name) if self.vararg_name else ""
        kwargs = ["%s=%s" % (n, v.__repr__()) for n, v in self.kwargs.items()]
        kw_wc = ("**" + self.kw_wildcard_name) if self.kw_wildcard_name else ""
        return "{name}({args}{c1}{vararg}{c2}{kwargs}{c3}{kw_wc})".format(
            name=self.name,
            args=", ".join(args),
            c1=", " if vararg and args else "",
            vararg=vararg,
            c2=", " if kwargs and (args or vararg) else "",
            kwargs=", ".join(kwargs),
            c3=", " if kw_wc and (args or vararg or kwargs) else "",
            kw_wc=kw_wc
        )

    def __repr__(self):
        return "<Signature at 0x{1:x} for '{0}'>".format(self.name, id(self))

    def _assert_no_unexpected_args(self, args):
        if self.vararg_name is not None:
            return
        if len(args) > len(self.arguments):
            unexpected_args = args[len(self.arguments):]
            raise TypeError("{} got unexpected argument(s): {}".format(
                self.name, unexpected_args))

    def _assert_no_unexpected_kwargs(self, kwargs):
        if self.kw_wildcard_name is not None:
            return
        unexpected_kwargs = [v for v in kwargs if v not in self.arguments]
        if unexpected_kwargs:
            raise TypeError("{} got unexpected kwarg(s): {}".format(
                self.name, unexpected_kwargs))

    def _assert_no_duplicate_args(self, args, kwargs):
        positional_arguments = self.arguments[:len(args)]
        duplicate_arguments = [v for v in positional_arguments if v in kwargs]
        if duplicate_arguments:
            raise TypeError("{} got multiple values for argument(s) {}".format(
                self.name, duplicate_arguments))

    def _fill_in_options(self, args, kwargs, options):
        free_params = self.get_free_parameters(args, kwargs)
        for f in free_params:
            if f in options:
                kwargs[f] = options[f]
        return args, kwargs

    def _assert_no_missing_args(self, args, kwargs):
        free_params = self.get_free_parameters(args, kwargs)
        missing_args = [m for m in free_params if m not in self.kwargs]
        if missing_args:
            raise TypeError("{} is missing value(s) for {}".format(
                self.name, missing_args))



#!/usr/bin/env python
# coding=utf-8
from __future__ import division, print_function, unicode_literals
import ast
from copy import copy
from functools import update_wrapper
import inspect
import json
import re
from sacred.custom_containers import dogmatize, undogmatize

try:
    import numpy as np
except ImportError:
    np = None


__sacred__ = True


def get_function_body_code(func):
    func_code_lines, start_idx = inspect.getsourcelines(func)
    filename = inspect.getfile(func)
    func_code = ''.join(func_code_lines)
    arg = "(?:[a-zA-Z_][a-zA-Z0-9_]*)"
    arguments = r"{0}(?:\s*,\s*{0})*".format(arg)
    func_def = re.compile(
        r"^[ \t]*def[ \t]*{}[ \t]*\(\s*({})?\s*\)[ \t]*:[ \t]*\n\s*".format(
            func.__name__, arguments), flags=re.MULTILINE)
    defs = list(re.finditer(func_def, func_code))
    assert defs
    line_offset = func_code[:defs[0].end()].count('\n')
    func_body = func_code[defs[0].end():]
    body_code = compile(inspect.cleandoc(func_body), filename, "exec",
                        ast.PyCF_ONLY_AST)
    body_code = ast.increment_lineno(body_code, n=start_idx+line_offset-1)
    body_code = compile(body_code, filename, "exec")
    return body_code


def recursive_fill_in(config, preset):
    for p in preset:
        if p not in config:
            config[p] = preset[p]
        elif isinstance(config[p], dict):
            recursive_fill_in(config[p], preset[p])


def chain_evaluate_config_scopes(config_scopes, fixed=None, preset=None,
                                 fallback=None):
    fixed = fixed or {}
    fallback = fallback or {}
    final_config = dict(preset or {})
    for config in config_scopes:
        config(fixed=fixed,
               preset=final_config,
               fallback=fallback)

        final_config.update(config)

    if not config_scopes:
        final_config.update(fixed)

    return undogmatize(final_config)


class ConfigScope(dict):
    def __init__(self, func):
        super(ConfigScope, self).__init__()
        self.arg_spec = inspect.getargspec(func)
        assert self.arg_spec.varargs is None, \
            "varargs are not allowed for ConfigScope functions"
        assert self.arg_spec.keywords is None, \
            "kwargs are not allowed for ConfigScope functions"
        assert self.arg_spec.defaults is None, \
            "default values are not allowed for ConfigScope functions"

        self._func = func
        update_wrapper(self, func)
        self._body_code = get_function_body_code(func)
        self._initialized = False
        self.added_values = set()
        self.typechanges = {}

    def __call__(self, fixed=None, preset=None, fallback=None):
        """
        Execute this ConfigScope. This will evaluate the function body and
        fill the relevant local variables into entries into keys in this
        dictionary.

        :param fixed: Dictionary of entries that should stay fixed during the
                      evaluation. All of them will be part of the final config.
        :type fixed: dict
        :param preset: Dictionary of preset values that will be available
                       during the evaluation (if they are declared in the
                       function argument list). All of them will be part of the
                       final config.
        :type preset: dict
        :param fallback: Dictionary of fallback values that will be available
                         during the evaluation (if they are declared in the
                         function argument list). They will NOT be part of the
                         final config.
        :type fallback: dict
        :return: self
        :rtype: ConfigScope
        """
        self._initialized = True
        self.clear()
        cfg_locals = dogmatize(fixed or {})
        fallback = fallback or {}
        preset = preset or {}
        fallback_view = {}

        available_entries = set(preset.keys()) | set(fallback.keys())

        for a in self.arg_spec.args:
            if a not in available_entries:
                raise KeyError("'%s' not in preset for ConfigScope. "
                               "Available options are: %s" %
                               (a, available_entries))
            if a in preset:
                cfg_locals[a] = preset[a]
            else:  # a in fallback
                fallback_view[a] = fallback[a]

        cfg_locals.fallback = fallback_view
        eval(self._body_code, copy(self._func.__globals__), cfg_locals)
        self.added_values = cfg_locals.revelation()
        self.typechanges = cfg_locals.typechanges

        # fill in the unused presets
        recursive_fill_in(cfg_locals, preset)

        for k, v in cfg_locals.items():
            if k.startswith('_'):
                continue
            if np and isinstance(v, np.bool_):
                # fixes an issue with numpy.bool_ not being json-serializable
                self[k] = bool(v)
                continue
            try:
                json.dumps(v)
                self[k] = undogmatize(v)
            except TypeError:
                pass
        return self

    def __getitem__(self, item):
        assert self._initialized, "ConfigScope must be executed before access"
        return dict.__getitem__(self, item)



#!/usr/bin/env python
# coding=utf-8

from __future__ import division, print_function, unicode_literals
from datetime import timedelta
import sys
import threading
import time
import traceback

from sacred.utils import tee_output


__sacred__ = True  # marker for filtering stacktraces when run from commandline


class Status(object):
    READY = 1
    RUNNING = 2
    COMPLETED = 3
    INTERRUPTED = 4
    FAILED = 5


class Run(object):
    """
    Represents and manages a single run of an experiment.
    """

    def __init__(self, config, config_modifications, main_function, observers,
                 logger, experiment_name, experiment_info, host_info):
        self.config = config
        self.main_function = main_function
        self.config_modifications = config_modifications
        self._observers = observers
        self.logger = logger
        self.experiment_name = experiment_name
        self.experiment_info = experiment_info
        self.host_info = host_info
        self.status = Status.READY
        self.info = {}
        self._heartbeat = None
        self.captured_out = None
        self.start_time = None
        self.stop_time = None
        self.elapsed_time = None
        self.result = None
        self._emit_run_created_event()

    def __call__(self, *args):
        with tee_output() as self.captured_out:
            self.logger.info('Started')
            self.status = Status.RUNNING
            self._emit_started()
            self._start_heartbeat()
            try:
                self.result = self.main_function(*args)
            except KeyboardInterrupt:
                self.status = Status.INTERRUPTED
                self._stop_heartbeat()
                self._emit_interrupted()
                raise
            except:
                self.status = Status.FAILED
                t, v, trace = sys.exc_info()
                self._stop_heartbeat()
                self._emit_failed(t, v, trace.tb_next)
                raise
            else:
                self.status = Status.COMPLETED
                self._stop_heartbeat()
                self._emit_completed(self.result)
                return self.result

    def _start_heartbeat(self):
        self._emit_heatbeat()
        self._heartbeat = threading.Timer(10, self._start_heartbeat)
        self._heartbeat.start()

    def _stop_heartbeat(self):
        if self._heartbeat is None:
            return
        self._heartbeat.cancel()
        self._heartbeat = None
        self._emit_heatbeat()  # one final beat to flush pending changes

    def _emit_run_created_event(self):
        for o in self._observers:
            try:
                o.created_event(**self.experiment_info)
            except AttributeError:
                pass

    def _emit_started(self):
        self.start_time = time.time()
        for o in self._observers:
            try:
                o.started_event(
                    name=self.experiment_name,
                    ex_info=self.experiment_info,
                    host_info=self.host_info,
                    start_time=self.start_time,
                    config=self.config)
            except AttributeError:
                pass

    def _emit_heatbeat(self):
        if self.status != Status.RUNNING:
            return

        for o in self._observers:
            try:
                o.heartbeat_event(
                    info=self.info,
                    captured_out=self.captured_out.getvalue())
            except AttributeError:
                pass

    def _stop_time(self):
        self.stop_time = time.time()
        elapsed_seconds = round(self.stop_time - self.start_time)
        self.elapsed_time = timedelta(seconds=elapsed_seconds)
        return self.stop_time

    def _emit_completed(self, result):
        stop_time = self._stop_time()
        self.logger.info('Completed after %s' % self.elapsed_time)
        for o in self._observers:
            try:
                o.completed_event(
                    stop_time=stop_time,
                    result=result)
            except AttributeError:
                pass

    def _emit_interrupted(self):
        interrupt_time = self._stop_time()
        self.logger.warning("Aborted after %s!" % self.elapsed_time)
        for o in self._observers:
            try:
                o.interrupted_event(
                    interrupt_time=interrupt_time)
            except AttributeError:
                pass

    def _emit_failed(self, etype, value, tb):
        fail_time = self._stop_time()
        self.logger.error("Failed after %s!" % self.elapsed_time)
        fail_trace = traceback.format_exception(etype, value, tb)
        for o in self._observers:
            try:
                o.failed_event(
                    fail_time=fail_time,
                    fail_trace=fail_trace)
            except:  # _emit_failed should never throw
                pass
